"""This module contains the base class for SQL steps."""

from typing import Any, Dict, Optional, Union
from abc import ABC
from pathlib import Path

from koheesio import Step
from koheesio.models import ExtraParamsMixin, Field, model_validator


class SqlBaseStep(Step, ExtraParamsMixin, ABC):
    """Base class for SQL steps

    `params` are used as placeholders for templating. The substitutions are identified by braces ('{' and '}') and can
    optionally contain a $-sign - e.g. `${placeholder}` or `{placeholder}`.

    Parameters
    ----------
    sql_path : Optional[Union[Path, str]], optional, default=None
        Path to a SQL file
    sql : Optional[str], optional, default=None
        SQL script to apply
    params : Dict[str, Any], optional, default_factory=dict
        Placeholders (parameters) for templating. These are identified with `${placeholder}` in the SQL script.\n
        __Note__: any arbitrary kwargs passed to the class will be added to params.
    """

    sql_path: Optional[Union[Path, str]] = Field(default=None, description="Path to a SQL file")
    sql: Optional[str] = Field(default=None, description="SQL script to apply")
    params: Dict[str, Any] = Field(
        default_factory=dict,
        description="Placeholders (parameters) for templating. The substitutions are identified by braces ('{' and '}')"
        "and can optionally contain a $-sign. Note: any arbitrary kwargs passed to the class will be added to params.",
    )

    @model_validator(mode="after")
    def _validate_sql_and_sql_path(self) -> "SqlBaseStep":
        """Validate the SQL and SQL path"""
        sql = self.sql
        sql_path = self.sql_path

        if sql_path is None and sql is None:
            raise ValueError("Please specify either `sql_path` or `sql`")

        if sql_path is not None and sql is not None:
            raise ValueError("You cannot specify both `sql_path` and `sql`")

        if sql_path is not None:
            sql_path = Path(sql_path)
            if not sql_path.exists():
                raise FileNotFoundError(f"Unable to locate specified {sql_path.as_posix()}")
            self.sql = sql_path.read_text(encoding="utf-8")
            return self

        if sql is None or len(sql.strip()) == 0:
            raise ValueError("SQL is empty")

        return self

    @property
    def query(self) -> str:
        """Returns the query while performing params replacement"""

        if self.sql:
            query = self.sql

            for key, value in self.params.items():
                query = query.replace(f"${{{key}}}", value)

            self.log.debug(f"Generated query: {query}")
        else:
            query = ""

        return query



"""
In this module you will find utility transformations that are useful for debugging and development purposes.
"""

from typing import Optional
from logging import DEBUG

from koheesio.models import Field
from koheesio.spark import SparkStep
from koheesio.spark.transformations import Transformation


class Peek(Transformation):
    """
    Peek the DataFrame: lets you inspect the content and/or the schema of the DataFrame at any given point of a Koheesio
    pipeline.

    You may want to run or disable this step depending on the environment the pipeline is running in. This can be
    accomplished by configuring the `enabled` and `enabled_by_logger` flags. The idea is that you can simply have the
    `enabled` flag decide whether to run the transformation (in this case, `enabled_by_logger` must be set to False),
    or you can run it based on the current log level. To accomplish that, set both `enabled` and `enabled_by_logger`
    to True: the step will be executed only if the log level is set to DEBUG.

    Here's a table that summarizes the scenarios and the expected behavior of the step:

    +----------------+---------------------+---------------------+---------------------+
    | enabled        | enabled_by_logger   | log level           | Is Peek enabled?    |
    +----------------+---------------------+---------------------+---------------------+
    | False          | *                   | *                   | NO                  |
    | True           | False               | *                   | YES                 |
    | True           | True                | DEBUG               | YES                 |
    | True           | True                | INFO or higher      | NO                  |
    +----------------+---------------------+---------------------+---------------------+

    You can simply slide in this transformation in between the transformations you wish to apply in order to peek the
    DataFrame data and schema:

    Example
    -------
    ```python
    from koheesio.spark.etl_task import EtlTask
    from koheesio.spark.readers.dummy import DummyReader
    from koheesio.spark.transformations.debug import Peek
    from koheesio.spark.transformations.transform import Transform
    from koheesio.spark.writers.dummy import DummyWriter

    reader = DummyReader(n=10)
    writer = DummyWriter()
    transformations = [
        Peek(enabled=True, enabled_by_logger=False, print_schema=True),
        Transform(
            lambda df: df.withColumn("id_squared", df["id"] * df["id"])
        ),
        Peek(enabled=True, enabled_by_logger=False, print_schema=True),
    ]
    task = EtlTask(
        source=reader, transformations=transformations, target=writer
    )
    task.execute()
    ```
    """

    enabled: bool = Field(
        ...,
        description="Whether to run this step or not. The resulting behavior may influenced by "
        "the enabled_by_logger paramter and the current log level. See the class "
        "docstring for more details.",
    )

    enabled_by_logger: Optional[bool] = Field(
        default=False,
        description="Enables the step if the log level of the logger is set to DEBUG or lower. The resulting behavior "
        "may influenced by the enabled parameter and the current log level. See the class docstring for "
        "more details.",
    )

    print_schema: Optional[bool] = Field(default=False, description="Enables printing the schema of the DataFrame")
    truncate_output: Optional[bool] = Field(
        default=False, description="Whether to truncate the output when printing the DataFrame"
    )
    vertical: Optional[bool] = Field(default=False, description="Print the DataFrame vertically")
    n: Optional[int] = Field(default=20, description="Number of rows to show")

    def execute(self) -> SparkStep.Output:
        if self.active:
            self._peek()

    def _peek(self):
        self.df = self.df.cache()

        if self.print_schema:
            self.df.printSchema()

        self.df.show(truncate=self.truncate_output, n=self.n, vertical=self.vertical)
        self.output.df = self.df

    @property
    def active(self) -> bool:
        # The is active parameter takes precedence over the logger level
        if not self.enabled:
            return False

        # Is active is True, and enabled by logger is False, we don't care about current log level
        if not self.enabled_by_logger:
            return True

        # If both flags are True, it will be active based on log level
        if self._on_debug():
            return True

        return False

    def _on_debug(self) -> bool:
        """
        Check if the logger level is set to DEBUG or lower.

        Returns
        --------
        bool
            True if the logger level is set to DEBUG or lower, False otherwise.
        """
        return self.log.isEnabledFor(DEBUG)



from typing import Generator
from io import StringIO
import logging

import pytest
from requests_mock.mocker import Mocker

from koheesio.logger import LoggingFactory
from koheesio.models import SecretStr
from koheesio.sso import okta as o

log = LoggingFactory.get_logger(name="test_download_file", inherit_from_koheesio=True)


@pytest.fixture(scope="function")
def requests_mocker() -> Generator[Mocker, None, None]:
    """Requests mocker fixture"""
    with Mocker() as rm:
        yield rm


# pylint: disable=assignment-from-no-return, redefined-outer-name
class TestOktaToken:
    """Tests for OktaAccessToken class"""

    url = "https://host.okta.com/oauth2/auth/v1/token"
    ot = o.OktaAccessToken(
        url=url,
        client_id="client",
        client_secret=SecretStr("secret"),
    )
    default_options = {
        "headers": {},
        "data": {"grant_type": "client_credentials"},
        "url": "url",
        "timeout": 3,
    }

    def test_okta_token_w_token(self, requests_mocker: Mocker) -> None:
        """Test OktaAccessToken with token"""
        # Arrange
        requests_mocker.post(self.url, json={"access_token": "bar"}, status_code=int(200))
        # Act
        output = self.ot.execute()
        # Assert
        assert output.token.get_secret_value() == "bar"

    def test_okta_token_wo_token(self, requests_mocker: Mocker) -> None:
        """Test OktaAccessToken without token"""
        requests_mocker.post(self.url, json={"foo": "bar"}, status_code=int(200))
        with pytest.raises(ValueError):
            self.ot.execute()

    def test_okta_token_non_200(self, requests_mocker: Mocker) -> None:
        """Test OktaAccessToken with non-200 response"""
        requests_mocker.post(self.url, status_code=int(404))
        with pytest.raises(o.HTTPError):
            self.ot.execute()

    def test_wo_extra_params(self) -> None:
        """Test OktaAccessToken without extra params"""
        oat = o.OktaAccessToken(url="url", client_id="client", client_secret=SecretStr("secret"))
        actual = oat.get_options()
        assert actual == {"auth": ("client", "secret"), **self.default_options}

    def test_w_extra_params(self) -> None:
        """Test OktaAccessToken with extra params"""
        oat = o.OktaAccessToken(url="url", client_id="client", client_secret=SecretStr("secret"), params={"foo": "bar"})
        assert oat.get_options() == {"foo": "bar", "auth": ("client", "secret"), **self.default_options}

    def test_log_extra_params_secret_happy(self, caplog: pytest.FixtureRequest) -> None:
        """Test that secret values are masked in logs"""
        # Arrange
        with caplog.at_level("DEBUG"):
            secret_val = "secret_value"
            # Act
            oat = o.OktaAccessToken(
                url="url", client_id="client", client_secret=SecretStr(secret_val), params={"foo": "bar"}
            )
            log.debug(f"{oat.get_options() = }")
            log.info(f"{oat.get_options() = }")
            log.warning(f"{oat.get_options() = }")
            log.error(f"{oat.get_options() = }")
            log.critical(f"{oat.get_options() = }")
            # Assert
            assert secret_val not in caplog.text
            assert "*" * len(secret_val) + "(Masked)" in caplog.text

    def test_secrets_not_in_log_unhappy(self, requests_mocker: Mocker, caplog: pytest.FixtureRequest) -> None:
        """Test that secret values are not logged upon errors"""
        # Arrange
        requests_mocker.post(self.url, status_code=int(404))
        with caplog.at_level("DEBUG"):
            secret_val = "unique_secret_value"
            # Act
            oat = o.OktaAccessToken(url=self.url, client_id="client", client_secret=SecretStr(secret_val))
            with pytest.raises(o.HTTPError):
                oat.execute()
            # Assert
            assert secret_val not in caplog.text

    def test_ensure_header_is_blank(self) -> None:
        """Test that the header is blank"""
        assert self.ot.headers == {}



import pytest

from koheesio.spark.readers.jdbc import JdbcReader

pytestmark = pytest.mark.spark


class TestJdbcReader:
    common_options = {
        "driver": "driver",
        "url": "url",
        "user": "user",
        "password": "password",
    }

    def test_get_options_wo_extra_options(self):
        jr = JdbcReader(**self.common_options, dbtable="table")
        actual = jr.get_options()
        del actual["password"]  # we don't need to test for this

        expected = {**self.common_options, "dbtable": "table"}
        del expected["password"]  # we don't need to test for this

        assert actual == expected

    def test_get_options_w_extra_options(self):
        jr = JdbcReader(
            options={
                "foo": "foo",
                "bar": "bar",
            },
            query="unit test",
            dbtable="table",
            **self.common_options,
        )

        actual = jr.get_options()
        del actual["password"]  # we don't need to test for this

        expected = {
            **self.common_options,
            "query": "unit test",
            "foo": "foo",
            "bar": "bar",
        }
        del expected["password"]  # we don't need to test for this

        assert sorted(actual) == sorted(expected)

        with pytest.raises(KeyError) as e:
            actual["dbtable"]
            assert e.type is KeyError

    def test_execute_wo_dbtable_and_query(self):
        with pytest.raises(ValueError) as e:
            _ = JdbcReader(**self.common_options)
            assert e.type is ValueError

    def test_execute_w_dbtable_and_query(self, dummy_spark):
        """query should take precedence over dbtable"""
        jr = JdbcReader(**self.common_options, dbtable="foo", query="bar")
        jr.execute()

        assert jr.df.count() == 3
        assert dummy_spark.options_dict["query"] == "bar"
        assert dummy_spark.options_dict.get("dbtable") is None

    def test_execute_w_dbtable(self, dummy_spark):
        """check that dbtable is passed to the reader correctly"""
        jr = JdbcReader(**self.common_options, dbtable="foo")
        jr.execute()

        assert jr.df.count() == 3
        assert dummy_spark.options_dict["dbtable"] == "foo"



"""
Excel reader for Spark

Note
----
Ensure the 'excel' extra is installed before using this reader.
Default implementation uses openpyxl as the engine for reading Excel files.
Other implementations can be used by passing the correct keyword arguments to the reader.

See Also
--------
- https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_excel.html
- koheesio.pandas.readers.excel.ExcelReader
"""

from typing import List, Optional, Union
from pathlib import Path

import pandas as pd

from koheesio.models import ExtraParamsMixin, Field
from koheesio.pandas.readers import Reader


class ExcelReader(Reader, ExtraParamsMixin):
    """Read data from an Excel file

    See Also
    --------
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_excel.html

    Attributes
    ----------
    path : Union[str, Path]
        The path to the Excel file
    sheet_name : str
        The name of the sheet to read
    header : Optional[Union[int, List[int]]]
        Row(s) to use as the column names

    Any other keyword arguments will be passed to pd.read_excel.
    """

    path: Union[str, Path] = Field(description="The path to the Excel file")
    sheet_name: str = Field(default="Sheet1", description="The name of the sheet to read")
    header: Optional[Union[int, List[int]]] = Field(default=0, description="Row(s) to use as the column names")

    def execute(self) -> Reader.Output:
        extra_params = self.params or {}
        extra_params.pop("spark", None)
        self.output.df = pd.read_excel(self.path, sheet_name=self.sheet_name, header=self.header, **extra_params)



from pyspark.sql import functions as f

from koheesio.spark import Column, SparkSession


def current_timestamp_utc(spark: SparkSession) -> Column:
    """Get the current timestamp in UTC"""
    tz_session = spark.conf.get("spark.sql.session.timeZone", "UTC")
    tz = tz_session if tz_session else "UTC"

    return f.to_utc_timestamp(f.current_timestamp(), tz)
